from __future__ import annotations

from typing import Any, Dict, Optional, Union

from cognite.client.utils._auxiliary import basic_obj_dump
from cognite.client.utils._text import convert_all_keys_to_snake_case, iterable_to_case


class TransformationDestination:
    """TransformationDestination has static methods to define the target resource type of a transformation

    Args:
        type (str): Used as data type identifier on transformation creation/retrieval.
    """

    def __init__(self, type: Optional[str] = None):
        self.type = type

    def __hash__(self) -> int:
        return hash(self.type)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and hash(other) == hash(self)

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        ret = basic_obj_dump(self, camel_case)

        needs_dump = set(iterable_to_case(("view", "edge_type", "data_model"), camel_case))
        for k in needs_dump.intersection(ret):
            if ret[k] is not None:
                ret[k] = ret[k].dump(camel_case=camel_case)
        return ret

    @staticmethod
    def assets() -> TransformationDestination:
        """To be used when the transformation is meant to produce assets."""
        return TransformationDestination(type="assets")

    @staticmethod
    def timeseries() -> TransformationDestination:
        """To be used when the transformation is meant to produce time series."""
        return TransformationDestination(type="timeseries")

    @staticmethod
    def asset_hierarchy() -> TransformationDestination:
        """To be used when the transformation is meant to produce asset hierarchies."""
        return TransformationDestination(type="asset_hierarchy")

    @staticmethod
    def events() -> TransformationDestination:
        """To be used when the transformation is meant to produce events."""
        return TransformationDestination(type="events")

    @staticmethod
    def datapoints() -> TransformationDestination:
        """To be used when the transformation is meant to produce numeric data points."""
        return TransformationDestination(type="datapoints")

    @staticmethod
    def string_datapoints() -> TransformationDestination:
        """To be used when the transformation is meant to produce string data points."""
        return TransformationDestination(type="string_datapoints")

    @staticmethod
    def sequences() -> TransformationDestination:
        """To be used when the transformation is meant to produce sequences."""
        return TransformationDestination(type="sequences")

    @staticmethod
    def files() -> TransformationDestination:
        """To be used when the transformation is meant to produce files."""
        return TransformationDestination(type="files")

    @staticmethod
    def labels() -> TransformationDestination:
        """To be used when the transformation is meant to produce labels."""
        return TransformationDestination(type="labels")

    @staticmethod
    def relationships() -> TransformationDestination:
        """To be used when the transformation is meant to produce relationships."""
        return TransformationDestination(type="relationships")

    @staticmethod
    def data_sets() -> TransformationDestination:
        """To be used when the transformation is meant to produce data sets."""
        return TransformationDestination(type="data_sets")

    @staticmethod
    def raw(database: str = "", table: str = "") -> RawTable:
        """To be used when the transformation is meant to produce raw table rows.

        Args:
            database (str): database name of the target raw table.
            table (str): name of the target raw table

        Returns:
            TransformationDestination pointing to the target table
        """
        return RawTable(database=database, table=table)

    @staticmethod
    def sequence_rows(external_id: str = "") -> SequenceRows:
        """To be used when the transformation is meant to produce sequence rows.

        Args:
            external_id (str): Sequence external id.

        Returns:
            TransformationDestination pointing to the target sequence rows
        """
        return SequenceRows(external_id=external_id)

    @staticmethod
    def nodes(view: Optional[ViewInfo] = None, instance_space: Optional[str] = None) -> Nodes:
        """

        Args:
            view (ViewInfo): information of the view.
            instance_space (str): space id of the instance.
        Returns:
            Nodes: pointing to the target flexible data model.
        """
        return Nodes(view=view, instance_space=instance_space)

    @staticmethod
    def edges(
        view: Optional[ViewInfo] = None,
        instance_space: Optional[str] = None,
        edge_type: Optional[EdgeType] = None,
    ) -> Edges:
        """

        Args:
            view (ViewInfo): information of the view.
            instance_space (str): space id of the instance.
            edge_type (EdgeType): information about the type of the edge
        Returns:
            Edges: pointing to the target flexible data model.
        """
        return Edges(view=view, instance_space=instance_space, edge_type=edge_type)

    @staticmethod
    def instances(data_model: Optional[DataModelInfo] = None, instance_space: Optional[str] = None) -> Instances:
        """
        Args:
            data_model (DataModelInfo): information of the Data Model.
            instance_space (str): space id of the instance.
        Returns:
            Instances: pointing to the target centric data model.
        """
        return Instances(data_model=data_model, instance_space=instance_space)


class RawTable(TransformationDestination):
    def __init__(self, database: Optional[str] = None, table: Optional[str] = None):
        super().__init__(type="raw")
        self.database = database
        self.table = table

    def __hash__(self) -> int:
        return hash((self.type, self.database, self.table))


class SequenceRows(TransformationDestination):
    def __init__(self, external_id: Optional[str] = None):
        super().__init__(type="sequence_rows")
        self.external_id = external_id

    def __hash__(self) -> int:
        return hash((self.type, self.external_id))


class ViewInfo:
    def __init__(self, space: str, external_id: str, version: str):
        self.space = space
        self.external_id = external_id
        self.version = version

    def __hash__(self) -> int:
        return hash((self.space, self.external_id, self.version))

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        return basic_obj_dump(self, camel_case)


class EdgeType:
    def __init__(self, space: str, external_id: str):
        self.space = space
        self.external_id = external_id

    def __hash__(self) -> int:
        return hash((self.space, self.external_id))

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        return basic_obj_dump(self, camel_case)


class DataModelInfo:
    def __init__(
        self,
        space: str,
        external_id: str,
        version: str,
        destination_type: str,
        destination_relationship_from_type: Optional[str] = None,
    ):

        self.space = space
        self.external_id = external_id
        self.version = version
        self.destination_type = destination_type
        self.destination_relationship_from_type = destination_relationship_from_type

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        return basic_obj_dump(self, camel_case)


class Nodes(TransformationDestination):
    def __init__(
        self,
        view: Optional[ViewInfo] = None,
        instance_space: Optional[str] = None,
    ):

        super().__init__(type="nodes")
        self.view = view
        self.instance_space = instance_space

    @classmethod
    def _load(cls, resource: Dict[str, Any]) -> Nodes:
        inst = cls(**resource)
        if isinstance(inst.view, dict):
            inst.view = ViewInfo(**convert_all_keys_to_snake_case(inst.view))
        return inst


class Edges(TransformationDestination):
    def __init__(
        self,
        view: Optional[ViewInfo] = None,
        instance_space: Optional[str] = None,
        edge_type: Optional[EdgeType] = None,
    ):

        super().__init__(type="edges")
        self.view = view
        self.instance_space = instance_space
        self.edge_type = edge_type

    @classmethod
    def _load(cls, resource: Dict[str, Any]) -> Edges:
        inst = cls(**resource)
        if isinstance(inst.view, dict):
            inst.view = ViewInfo(**convert_all_keys_to_snake_case(inst.view))
        if isinstance(inst.edge_type, dict):
            inst.edge_type = EdgeType(**convert_all_keys_to_snake_case(inst.edge_type))
        return inst


class Instances(TransformationDestination):
    def __init__(
        self,
        data_model: Optional[DataModelInfo] = None,
        instance_space: Optional[str] = None,
    ):
        super().__init__(type="instances")
        self.data_model = data_model
        self.instance_space = instance_space

    @classmethod
    def _load(cls, resource: Dict[str, Any]) -> Instances:
        inst = cls(**resource)
        if isinstance(inst.data_model, dict):
            inst.data_model = DataModelInfo(**convert_all_keys_to_snake_case(inst.data_model))
        return inst


class OidcCredentials:
    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        scopes: Optional[str] = None,
        token_uri: Optional[str] = None,
        audience: Optional[str] = None,
        cdf_project_name: Optional[str] = None,
    ):

        self.client_id = client_id
        self.client_secret = client_secret
        self.scopes = scopes
        self.token_uri = token_uri
        self.audience = audience
        self.cdf_project_name = cdf_project_name

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance.
        """
        return basic_obj_dump(self, camel_case)


class NonceCredentials:
    def __init__(
        self,
        session_id: int,
        nonce: str,
        cdf_project_name: str,
    ):
        self.session_id = session_id
        self.nonce = nonce
        self.cdf_project_name = cdf_project_name

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance.
        """
        return basic_obj_dump(self, camel_case)


class TransformationBlockedInfo:
    """Information about the reason why and when a transformation is blocked.

    Args:
        reason (str): Reason why the transformation is blocked.
        created_time (Optional[int]): Timestamp when the transformation was blocked.
    """

    def __init__(self, reason: Optional[str] = None, created_time: Optional[int] = None):
        self.reason = reason
        self.created_time = created_time


def _load_destination_dct(
    dct: Dict[str, Any]
) -> Union[RawTable, Nodes, Edges, SequenceRows, TransformationDestination]:
    """Helper function to load destination from dictionary"""
    snake_dict = convert_all_keys_to_snake_case(dct)
    destination_type = snake_dict.pop("type")
    simple = {
        "raw": RawTable,
        "sequence_rows": SequenceRows,
    }
    if destination_type in simple:
        return simple[destination_type](**snake_dict)

    nested: Dict[str, type[Nodes] | type[Edges] | type[Instances]] = {
        "nodes": Nodes,
        "edges": Edges,
        "instances": Instances,
    }
    if destination_type in nested:
        return nested[destination_type]._load(snake_dict)

    return TransformationDestination(destination_type)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class NumericDatapoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    timestamp: builtins.int
    value: builtins.float
    def __init__(
        self,
        *,
        timestamp: builtins.int = ...,
        value: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["timestamp", b"timestamp", "value", b"value"]) -> None: ...

global___NumericDatapoint = NumericDatapoint

@typing_extensions.final
class NumericDatapoints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAPOINTS_FIELD_NUMBER: builtins.int
    @property
    def datapoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___NumericDatapoint]: ...
    def __init__(
        self,
        *,
        datapoints: collections.abc.Iterable[global___NumericDatapoint] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["datapoints", b"datapoints"]) -> None: ...

global___NumericDatapoints = NumericDatapoints

@typing_extensions.final
class StringDatapoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    timestamp: builtins.int
    value: builtins.str
    def __init__(
        self,
        *,
        timestamp: builtins.int = ...,
        value: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["timestamp", b"timestamp", "value", b"value"]) -> None: ...

global___StringDatapoint = StringDatapoint

@typing_extensions.final
class StringDatapoints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAPOINTS_FIELD_NUMBER: builtins.int
    @property
    def datapoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StringDatapoint]: ...
    def __init__(
        self,
        *,
        datapoints: collections.abc.Iterable[global___StringDatapoint] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["datapoints", b"datapoints"]) -> None: ...

global___StringDatapoints = StringDatapoints

@typing_extensions.final
class AggregateDatapoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_FIELD_NUMBER: builtins.int
    AVERAGE_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    COUNT_FIELD_NUMBER: builtins.int
    SUM_FIELD_NUMBER: builtins.int
    INTERPOLATION_FIELD_NUMBER: builtins.int
    STEPINTERPOLATION_FIELD_NUMBER: builtins.int
    CONTINUOUSVARIANCE_FIELD_NUMBER: builtins.int
    DISCRETEVARIANCE_FIELD_NUMBER: builtins.int
    TOTALVARIATION_FIELD_NUMBER: builtins.int
    timestamp: builtins.int
    average: builtins.float
    max: builtins.float
    min: builtins.float
    count: builtins.float
    sum: builtins.float
    interpolation: builtins.float
    stepInterpolation: builtins.float
    continuousVariance: builtins.float
    discreteVariance: builtins.float
    totalVariation: builtins.float
    def __init__(
        self,
        *,
        timestamp: builtins.int = ...,
        average: builtins.float = ...,
        max: builtins.float = ...,
        min: builtins.float = ...,
        count: builtins.float = ...,
        sum: builtins.float = ...,
        interpolation: builtins.float = ...,
        stepInterpolation: builtins.float = ...,
        continuousVariance: builtins.float = ...,
        discreteVariance: builtins.float = ...,
        totalVariation: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["average", b"average", "continuousVariance", b"continuousVariance", "count", b"count", "discreteVariance", b"discreteVariance", "interpolation", b"interpolation", "max", b"max", "min", b"min", "stepInterpolation", b"stepInterpolation", "sum", b"sum", "timestamp", b"timestamp", "totalVariation", b"totalVariation"]) -> None: ...

global___AggregateDatapoint = AggregateDatapoint

@typing_extensions.final
class AggregateDatapoints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAPOINTS_FIELD_NUMBER: builtins.int
    @property
    def datapoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___AggregateDatapoint]: ...
    def __init__(
        self,
        *,
        datapoints: collections.abc.Iterable[global___AggregateDatapoint] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["datapoints", b"datapoints"]) -> None: ...

global___AggregateDatapoints = AggregateDatapoints



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, cast

from cognite.client.data_classes._base import CogniteFilter, CogniteResource, CogniteResourceList

if TYPE_CHECKING:
    from cognite.client import CogniteClient


class TransformationNotification(CogniteResource):
    """The transformation notification resource allows configuring email alerts on events related to a transformation run.

    Args:
        id (int): A server-generated ID for the object.
        transformation_id (int): Transformation Id.
        transformation_external_id (str): Transformation external Id.
        destination (str): Email address where notifications should be sent.
        created_time (int): Time when the notification was created.
        last_updated_time (int): Time when the notification was last updated.
        cognite_client (CogniteClient): The client to associate with this object.
    """

    def __init__(
        self,
        id: Optional[int] = None,
        transformation_id: Optional[int] = None,
        transformation_external_id: Optional[str] = None,
        destination: Optional[str] = None,
        created_time: Optional[int] = None,
        last_updated_time: Optional[int] = None,
        cognite_client: Optional[CogniteClient] = None,
    ):
        self.id = id
        self.transformation_id = transformation_id
        self.transformation_external_id = transformation_external_id
        self.destination = destination
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self._cognite_client = cast("CogniteClient", cognite_client)

    def __hash__(self) -> int:
        return hash(self.id)


class TransformationNotificationList(CogniteResourceList[TransformationNotification]):
    _RESOURCE = TransformationNotification


class TransformationNotificationFilter(CogniteFilter):
    """

    Args:
        transformation_id (Optional[int]): Filter by transformation internal numeric ID.
        transformation_external_id (str): Filter by transformation externalId.
        destination (str): Filter by notification destination.
    """

    def __init__(
        self,
        transformation_id: Optional[int] = None,
        transformation_external_id: Optional[str] = None,
        destination: Optional[str] = None,
    ):
        self.transformation_id = transformation_id
        self.transformation_external_id = transformation_external_id
        self.destination = destination



from __future__ import annotations

from typing import Any, Optional

from cognite.client.data_classes._base import (
    CogniteResourceList,
)
from cognite.client.data_classes.data_modeling._core import DataModelingResource
from cognite.client.data_classes.data_modeling._validation import validate_data_modeling_identifier


class SpaceCore(DataModelingResource):
    """A workspace for data models and instances.

    Args:
        space (str): A unique identifier for space.
        description (str): Textual description of the space
        name (str): Human readable name for the space.
    """

    def __init__(self, space: str, description: Optional[str] = None, name: Optional[str] = None):
        self.space = space
        self.description = description
        self.name = name

    def as_id(self) -> str:
        return self.space


class SpaceApply(SpaceCore):
    """A workspace for data models and instances. This is the write version

    Args:
        space (str): A unique identifier for space.
        description (str): Textual description of the space
        name (str): Human readable name for the space.
    """

    def __init__(
        self,
        space: str,
        description: Optional[str] = None,
        name: Optional[str] = None,
        **_: Any,
    ):
        validate_data_modeling_identifier(space)
        super().__init__(space, description, name)


class Space(SpaceCore):
    """A workspace for data models and instances. This is the read version.

    Args:
        space (str): a unique identifier for the space.
        description (str): Textual description of the space
        name (str): Human readable name for the space.
        is_global (bool): Whether the space is global or not.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
    """

    def __init__(
        self,
        space: str,
        is_global: bool,
        last_updated_time: int,
        created_time: int,
        description: Optional[str] = None,
        name: Optional[str] = None,
        **_: Any,
    ):
        super().__init__(space, description, name)
        self.is_global = is_global
        self.last_updated_time = last_updated_time
        self.created_time = created_time

    def as_apply(self) -> SpaceApply:
        return SpaceApply(
            space=self.space,
            description=self.description,
            name=self.name,
        )


class SpaceApplyList(CogniteResourceList[SpaceApply]):
    _RESOURCE = SpaceApply


class SpaceList(CogniteResourceList[Space]):
    _RESOURCE = Space

    def to_space_apply_list(self) -> SpaceApplyList:
        return SpaceApplyList(
            resources=[item.as_apply() for item in self.items],
            cognite_client=self._cognite_client,
        )



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, Type

from cognite.client.data_classes import Datapoints, filters
from cognite.client.data_classes._base import (
    CogniteListUpdate,
    CognitePrimitiveUpdate,
    CogniteResource,
    CogniteResourceList,
    CogniteUpdate,
    PropertySpec,
    T_CogniteResource,
)
from cognite.client.data_classes.filters import Filter
from cognite.client.utils._auxiliary import exactly_one_is_not_none
from cognite.client.utils._text import convert_all_keys_to_snake_case

if TYPE_CHECKING:
    from cognite.client import CogniteClient

ExternalId = str

_FILTERS_SUPPORTED: set[type[Filter]] = {
    filters.And,
    filters.Or,
    filters.Not,
    filters.In,
    filters.Equals,
    filters.Exists,
    filters.Range,
    filters.Prefix,
    filters.ContainsAny,
    filters.ContainsAll,
}


def _validate_filter(filter: Filter | None) -> None:
    if filter is None:
        return
    if not_supported := (filter._involved_filter_types() - _FILTERS_SUPPORTED):
        names = [f.__name__ for f in not_supported]
        raise ValueError(f"The filters {names} are not supported for DataPointSubscriptions")


class DatapointSubscriptionCore(CogniteResource):
    def __init__(
        self,
        external_id: ExternalId,
        partition_count: int,
        filter: Optional[Filter] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        **_: dict,
    ):
        self.external_id = external_id
        self.partition_count = partition_count
        self.filter = filter
        self.name = name
        self.description = description

    @classmethod
    def _load(
        cls: Type[T_CogniteResource], resource: dict | str, cognite_client: Optional[CogniteClient] = None
    ) -> T_CogniteResource:
        resource = json.loads(resource) if isinstance(resource, str) else resource
        if "filter" in resource:
            resource["filter"] = Filter.load(resource["filter"])

        resource = convert_all_keys_to_snake_case(resource)
        return cls(**resource)

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        data = super().dump(camel_case)
        if "filter" in data:
            data["filter"] = data["filter"].dump()
        return data


class DatapointSubscription(DatapointSubscriptionCore):
    """A data point subscription is a way to listen to changes to time series data points, in ingestion order.
        This is the read version of a subscription, used when reading subscriptions from CDF.

    Args:
        external_id (str): Externally provided ID for the subscription. Must be unique.
        partition_count (int): The maximum effective parallelism of this subscription (the number of clients that can
                               read from it concurrently) will be limited to this number, but a higher partition count
                               will cause a higher time overhead.
        created_time (int): Time when the subscription was created in CDF in milliseconds since Jan 1, 1970.
        last_updated_time (int): Time when the subscription was last updated in CDF in milliseconds since Jan 1, 1970.
        time_series_count (int): The number of time series in the subscription.
        filter (Optional[Filter]): If present, the subscription is defined by this filter.
        name (Optional[str]) Human readable name of the subscription.
        description (Optional[str]): A summary explanation for the subscription.
    """

    def __init__(
        self,
        external_id: ExternalId,
        partition_count: int,
        created_time: int,
        last_updated_time: int,
        time_series_count: int,
        filter: Optional[Filter] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        **_: dict,
    ):
        super().__init__(external_id, partition_count, filter, name, description)
        self.time_series_count = time_series_count
        self.created_time = created_time
        self.last_updated_time = last_updated_time


class DataPointSubscriptionCreate(DatapointSubscriptionCore):
    """A data point subscription is a way to listen to changes to time series data points, in ingestion order.
        This is the write version of a subscription, used to create new subscriptions.

    A subscription can either be defined directly by a list of time series ids or indirectly by a filter.

    Args:
        external_id (str): Externally provided ID for the subscription. Must be unique.
        partition_count (int): The maximum effective parallelism of this subscription (the number of clients that can
                               read from it concurrently) will be limited to this number, but a higher partition count
                               will cause a higher time overhead. The partition count must be between 1 and 100.
                               CAVEAT: This cannot change after the subscription has been created.
        time_series_ids (Optional[list[str]]): List of (external) ids of time series that this subscription will listen to.
                                     Not compatible with filter.
        filter (Optional[Filter]): A filter DSL (Domain Specific Language) to define advanced filter queries. Not compatible
                         with time_series_ids.
        name (Optional[str]) Human readable name of the subscription.
        description (Optional[str]): A summary explanation for the subscription.
    """

    def __init__(
        self,
        external_id: str,
        partition_count: int,
        time_series_ids: Optional[list[ExternalId]] = None,
        filter: Optional[Filter] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        if not exactly_one_is_not_none(time_series_ids, filter):
            raise ValueError("Exactly one of time_series_ids and filter must be given")
        _validate_filter(filter)
        super().__init__(external_id, partition_count, filter, name, description)
        self.time_series_ids = time_series_ids


class DataPointSubscriptionUpdate(CogniteUpdate):
    """Changes applied to datapoint subscription

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
    """

    def __init__(self, external_id: str):
        super().__init__(external_id=external_id)

    class _PrimitiveDataPointSubscriptionUpdate(CognitePrimitiveUpdate):
        def set(self, value: Any) -> DataPointSubscriptionUpdate:
            return self._set(value)

    class _FilterDataPointSubscriptionUpdate(CognitePrimitiveUpdate):
        def set(self, value: Filter) -> DataPointSubscriptionUpdate:
            return self._set(value.dump())  # type: ignore[arg-type]

    class _ListDataPointSubscriptionUpdate(CogniteListUpdate):
        def set(self, value: list) -> DataPointSubscriptionUpdate:
            return self._set(value)

        def add(self, value: list) -> DataPointSubscriptionUpdate:
            return self._add(value)

        def remove(self, value: list) -> DataPointSubscriptionUpdate:
            return self._remove(value)

    @property
    def name(self) -> _PrimitiveDataPointSubscriptionUpdate:
        return DataPointSubscriptionUpdate._PrimitiveDataPointSubscriptionUpdate(self, "name")

    @property
    def time_series_ids(self) -> _ListDataPointSubscriptionUpdate:
        return DataPointSubscriptionUpdate._ListDataPointSubscriptionUpdate(self, "timeSeriesIds")

    @property
    def filter(self) -> _FilterDataPointSubscriptionUpdate:
        return DataPointSubscriptionUpdate._FilterDataPointSubscriptionUpdate(self, "filter")

    @classmethod
    def _get_update_properties(cls) -> list[PropertySpec]:
        return [
            PropertySpec("name"),
            PropertySpec("time_series_ids", is_container=True),
            PropertySpec("filter", is_nullable=False),
        ]


class TimeSeriesID(CogniteResource):
    """
    A TimeSeries Identifier to uniquely identify a time series.

    Args:
        id (int): A server-generated ID for the object.
        external_id (Optional[str]): The external ID provided by the client. Must be unique for the resource type.
    """

    def __init__(self, id: int, external_id: ExternalId | None = None):
        self.id = id
        self.external_id = external_id

    @classmethod
    def _load(cls, resource: dict | str, cognite_client: Optional[CogniteClient] = None) -> TimeSeriesID:
        resource = json.loads(resource) if isinstance(resource, str) else resource
        return cls(id=resource["id"], external_id=resource.get("externalId"))

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        resource: dict[str, Any] = {"id": self.id}
        if self.external_id is not None:
            resource["externalId" if camel_case else "external_id"] = self.external_id
        return resource


@dataclass
class DataDeletion:
    inclusive_begin: int
    exclusive_end: int | None

    @classmethod
    def _load(cls, data: dict[str, Any]) -> DataDeletion:
        return cls(inclusive_begin=data["inclusiveBegin"], exclusive_end=data.get("exclusiveEnd"))

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        resource: dict[str, Any] = {("inclusiveBegin" if camel_case else "inclusive_begin"): self.inclusive_begin}
        if self.exclusive_end is not None:
            resource["exclusiveEnd" if camel_case else "exclusive_end"] = self.exclusive_end
        return resource


@dataclass
class DatapointsUpdate:
    time_series: TimeSeriesID
    upserts: Datapoints
    deletes: list[DataDeletion]

    @classmethod
    def _load(cls, data: dict[str, Any]) -> DatapointsUpdate:
        datapoints: dict[str, Any] = {"upserts": Datapoints(), "deletes": []}
        if (values := data["upserts"]) and ("value" in values[0]):
            datapoints["upserts"] = Datapoints._load(
                {
                    "id": data["timeSeries"]["id"],
                    "externalId": data["timeSeries"].get("externalId"),
                    "isString": isinstance(values[0]["value"], str),
                    "datapoints": values,
                }
            )
        if values := data["deletes"]:
            datapoints["deletes"] = [DataDeletion._load(value) for value in values]
        return cls(
            time_series=TimeSeriesID._load(data["timeSeries"], None),
            **datapoints,
        )

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        resource: dict[str, Any] = {("timeSeries" if camel_case else "time_series"): self.time_series.dump(camel_case)}
        if self.upserts is not None:
            resource["upserts"] = self.upserts.dump(camel_case)
        if self.deletes is not None:
            resource["deletes"] = [d.dump(camel_case) for d in self.deletes]
        return resource


@dataclass
class SubscriptionTimeSeriesUpdate:
    added: list[TimeSeriesID]
    removed: list[TimeSeriesID]

    @classmethod
    def _load(cls, data: dict[str, Any]) -> SubscriptionTimeSeriesUpdate:
        return cls(
            added=[TimeSeriesID._load(added) for added in data.get("added", [])],
            removed=[TimeSeriesID._load(added) for added in data.get("removed", [])],
        )

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        resource: dict[str, Any] = {}
        resource["added"] = [id_.dump() for id_ in self.added]
        resource["removed"] = [id_.dump() for id_ in self.removed]
        return resource


@dataclass
class DatapointSubscriptionPartition:
    index: int
    cursor: str | None = None

    @classmethod
    def create(cls, data: tuple[int, str] | int | DatapointSubscriptionPartition) -> DatapointSubscriptionPartition:
        if isinstance(data, DatapointSubscriptionPartition):
            return data
        if isinstance(data, tuple):
            return cls(*data)
        return cls(data)

    @classmethod
    def _load(cls, data: dict[str, Any]) -> DatapointSubscriptionPartition:
        return cls(index=data["index"], cursor=data.get("cursor") or data.get("nextCursor"))

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        output: dict[str, Any] = {"index": self.index}
        if self.cursor is not None:
            output["cursor"] = self.cursor
        return output


class _DatapointSubscriptionBatch:
    """A batch of data from a subscription.

    Args:
        updates (list[DatapointsUpdate]): List of updates from the subscription, sorted by point in time they were
                                         applied to the time series. Every update contains a time series along with a
                                         set of changes to that time series.
        partitions (list[DatapointSubscriptionPartition]): Which partitions/cursors to use for the next request.
                                                     Map from partition index to cursor.
        has_next (list[str): Whether there is more data available at the time of the query. In rare cases, we may
                             return true, even if there is no data available. If that is the case, just continue to
                             query with the updated cursors, and it will eventually return false.
        subscription_changes (SubscriptionTimeSeriesUpdate): If present, this object represents changes to the subscription
                                                 definition. The subscription will now start/stop listening to changes
                                                 from the time series listed here.
    """

    def __init__(
        self,
        updates: list[DatapointsUpdate],
        partitions: list[DatapointSubscriptionPartition],
        has_next: bool,
        subscription_changes: SubscriptionTimeSeriesUpdate,
    ):
        self.updates = updates
        self.partitions = partitions
        self.has_next = has_next
        self.subscription_changes = subscription_changes

    @classmethod
    def _load(cls, resource: dict | str) -> _DatapointSubscriptionBatch:
        resource = json.loads(resource) if isinstance(resource, str) else resource
        return cls(
            updates=[DatapointsUpdate._load(u) for u in resource["updates"]],
            partitions=[DatapointSubscriptionPartition._load(p) for p in resource["partitions"]],
            has_next=resource["hasNext"],
            subscription_changes=SubscriptionTimeSeriesUpdate._load(resource.get("subscriptionChanges", [])),
        )

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        resource: dict[str, Any] = {
            "updates": [u.dump(camel_case) for u in self.updates],
            "partitions": [p.dump(camel_case) for p in self.partitions],
            ("hasNext" if camel_case else "has_next"): self.has_next,
        }
        if self.subscription_changes is not None:
            resource[
                ("subscriptionChanges" if camel_case else "subscription_changes")
            ] = self.subscription_changes.dump(camel_case)
        return resource


class DatapointSubscriptionList(CogniteResourceList[DatapointSubscription]):
    _RESOURCE = DatapointSubscription


def _metadata(key: str) -> list[str]:
    return ["metadata", key]


class DatapointSubscriptionFilterProperties:
    description = ["description"]
    external_id = ["externalId"]
    metadata = _metadata
    name = ["name"]
    unit = ["unit"]
    asset_id = ["assetId"]
    asset_root_id = ["assetRootId"]
    created_time = ["createdTime"]
    data_set_id = ["dataSetId"]
    id = ["id"]
    last_updated_time = ["lastUpdatedTime"]
    is_step = ["isStep"]
    is_string = ["isString"]



from __future__ import annotations

LIST_LIMIT_DEFAULT = 25
# Max JavaScript-safe integer 2^53 - 1
MAX_VALID_INTERNAL_ID = 9007199254740991
DATA_MODELING_LIST_LIMIT_DEFAULT = 10
INSTANCES_LIST_LIMIT_DEFAULT = 1_000
DATAPOINT_SUBSCRIPTIONS_LIST_LIMIT_DEFAULT = 100
DATAPOINT_SUBSCRIPTION_DATA_LIST_LIMIT_DEFAULT = 100



"""
This code has been heavily modified from the original created by Oleg Lupats, 2019, under an MIT license:
project = 'PriorityThreadPoolExecutor'
url = 'https://github.com/oleglpts/PriorityThreadPoolExecutor'
copyright = '2019, Oleg Lupats'
author = 'Oleg Lupats'
release = '0.0.1'

MIT License

Copyright (c) 2019 The Python Packaging Authority

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import inspect
import itertools
import sys
import weakref
from concurrent.futures import Future, as_completed
from concurrent.futures.thread import ThreadPoolExecutor, _base, _WorkItem
from queue import Empty, PriorityQueue
from threading import Lock, Thread
from typing import Iterable, Iterator

NULL_ENTRY = (-1, None, None)
_THREADS_QUEUES = weakref.WeakKeyDictionary()
_SHUTDOWN = False
# Lock that ensures that new workers are not created while the interpreter is
# shutting down. Must be held while mutating _THREADS_QUEUES and _SHUTDOWN.
_GLOBAL_SHUTDOWN_LOCK = Lock()


def python_exit():
    global _SHUTDOWN
    with _GLOBAL_SHUTDOWN_LOCK:
        _SHUTDOWN = True
    items = list(_THREADS_QUEUES.items())
    for thread, queue in items:
        queue.put(NULL_ENTRY)
    for thread, queue in items:
        thread.join()


# Starting with 3.9, ThreadPoolExecutor no longer uses daemon threads, and so instead, an internal
# function hook very similar to atexit.register() gets called at 'threading' shutdown instead of
# interpreter shutdown. Check out https://github.com/python/cpython/issues/83993
if sys.version_info[:2] < (3, 9):
    from atexit import register as _register_atexit
else:
    from threading import _register_atexit

_register_atexit(python_exit)


def _worker(executor_reference, work_queue):
    try:
        while True:
            work_item = work_queue.get(block=True)[-1]
            if work_item is not None:
                work_item.run()
                del work_item
                continue
            executor = executor_reference()
            if _SHUTDOWN or executor is None or executor._shutdown:
                work_queue.put(NULL_ENTRY)
                return None
            del executor
    except BaseException:
        _base.LOGGER.critical("Exception in worker", exc_info=True)


class PriorityThreadPoolExecutor(ThreadPoolExecutor):
    """Thread pool executor with queue.PriorityQueue() as its work queue. Accepts a 'priority' parameter
    thats controls the prioritisation of tasks: lower numbers being run before higher numbers, and
    0 (zero) being the highest possible priority.

    All tasks not given a priority will be given `priority=0` to work seamlessly as a stand-in for the
    regular ThreadPoolExecutor.
    """

    def __init__(self, max_workers=None):
        super().__init__(max_workers)
        self._work_queue = PriorityQueue()
        self._task_counter = itertools.count().__next__

    @staticmethod
    def as_completed(futures: Iterable[Future]) -> Iterator[Future]:
        # This is just here to make a serial non-threading version "hot-swappable"
        return as_completed(futures)

    def submit(self, fn, *args, **kwargs):
        if "priority" in inspect.signature(fn).parameters:
            raise TypeError(f"Given function {fn} cannot accept reserved parameter name `priority`")

        with self._shutdown_lock, _GLOBAL_SHUTDOWN_LOCK:
            if self._shutdown:
                raise RuntimeError("Cannot schedule new futures after shutdown")

            if _SHUTDOWN:
                raise RuntimeError("Cannot schedule new futures after interpreter shutdown")

            if (priority := kwargs.pop("priority", 0)) < 0:
                raise ValueError("'priority' has to be a nonnegative number, 0 being the highest possible priority")

            future = Future()
            work_item = _WorkItem(future, fn, args, kwargs)

            # We use a counter to break ties, but keep order:
            self._work_queue.put((priority, self._task_counter(), work_item))
            self._adjust_thread_count()
            return future

    def _adjust_thread_count(self):
        def weak_ref_cb(_, queue=self._work_queue):
            queue.put(NULL_ENTRY)

        if len(self._threads) < self._max_workers:
            thread = Thread(target=_worker, args=(weakref.ref(self, weak_ref_cb), self._work_queue))
            thread.daemon = True
            thread.start()
            self._threads.add(thread)
            _THREADS_QUEUES[thread] = self._work_queue

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            self._shutdown = True

            # Empty the entire work queue: (mod. from py39's added option 'cancel_futures')
            while True:
                try:
                    work_item = self._work_queue.get_nowait()[-1]
                except Empty:
                    break
                if work_item is not None:
                    work_item.future.cancel()

            # Send a wake-up to prevent threads calling _work_queue.get(block=True) from permanently blocking
            self._work_queue.put(NULL_ENTRY)

        if wait:
            for thread in self._threads:
                thread.join()



from concurrent.futures import ThreadPoolExecutor

from cognite.client.utils._concurrency import ConcurrencySettings, MainThreadExecutor, TaskExecutor

_THREAD_POOL_EXECUTOR_SINGLETON: ThreadPoolExecutor
_MAIN_THREAD_EXECUTOR_SINGLETON = MainThreadExecutor()
_MAX_WORKERS = 2


def get_data_modeling_executor() -> TaskExecutor:
    """
    The data modeling backend has different concurrency limits in the backend compared to the rest of CDF.
    Thus, we use a dedicated executor for these endpoints to match the backend.

    Returns:
        The data modeling executor.
    """
    global _THREAD_POOL_EXECUTOR_SINGLETON

    if ConcurrencySettings.executor_type == "threadpool":
        try:
            executor: TaskExecutor = _THREAD_POOL_EXECUTOR_SINGLETON
        except NameError:
            # TPE has not been initialized
            executor = _THREAD_POOL_EXECUTOR_SINGLETON = ThreadPoolExecutor(_MAX_WORKERS)
    elif ConcurrencySettings.executor_type == "mainthread":
        executor = _MAIN_THREAD_EXECUTOR_SINGLETON
    else:
        raise RuntimeError(f"Invalid executor type '{ConcurrencySettings.executor_type}'")
    return executor



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class DMLApplyResult:
    space: str
    external_id: str
    version: str
    name: Optional[str]
    description: Optional[str]
    created_time: int
    last_updated_time: int

    @classmethod
    def load(cls, data: dict[str, Any]) -> DMLApplyResult:
        return cls(
            space=data["space"],
            external_id=data["externalId"],
            version=data["version"],
            name=data["name"],
            description=data["description"],
            created_time=data["createdTime"],
            last_updated_time=data["lastUpdatedTime"],
        )



from __future__ import annotations

import argparse
import functools
import re
from typing import List, Optional, Tuple

import requests

from cognite.client.config import global_config


def check_if_version_exists(package_name: str, version: str) -> bool:
    versions = get_all_versions(package_name=package_name)
    return version in versions


@functools.lru_cache(maxsize=1)
def get_newest_version_in_major_release(package_name: str, version: str) -> str:
    major, minor, micro, pr_cycle, pr_version = _parse_version(version)
    versions = get_all_versions(package_name)
    for v in versions:
        if _is_newer_major(v, version):
            major, minor, micro, pr_cycle, pr_version = _parse_version(v)
    return _format_version(major, minor, micro, pr_cycle, pr_version)


def get_all_versions(package_name: str) -> List[str]:
    verify_ssl = not global_config.disable_ssl
    res = requests.get(f"https://pypi.python.org/simple/{package_name}/#history", verify=verify_ssl, timeout=5)
    return re.findall(r"cognite-sdk-(\d+\.\d+.[\dabrc]+)", res.content.decode())


def _is_newer_major(version_a: str, version_b: str) -> bool:
    major_a, minor_a, micro_a, pr_cycle_a, pr_version_a = _parse_version(version_a)
    major_b, minor_b, micro_b, pr_cycle_b, pr_version_b = _parse_version(version_b)
    is_newer = False
    if major_a == major_b and minor_a >= minor_b and micro_a >= micro_b:
        if minor_a > minor_b:
            is_newer = True
        else:
            if micro_a > micro_b:
                is_newer = True
            else:
                is_newer = _is_newer_pre_release(pr_cycle_a, pr_version_a, pr_cycle_b, pr_version_b)
    return is_newer


def _is_newer_pre_release(
    pr_cycle_a: Optional[str], pr_v_a: Optional[int], pr_cycle_b: Optional[str], pr_v_b: Optional[int]
) -> bool:
    cycles = ["a", "b", "rc", None]
    assert pr_cycle_a in cycles, f"pr_cycle_a must be one of '{pr_cycle_a}', not '{cycles}'."
    assert pr_cycle_b in cycles, f"pr_cycle_a must be one of '{pr_cycle_b}', not '{cycles}'."
    is_newer = False
    if cycles.index(pr_cycle_a) > cycles.index(pr_cycle_b):
        is_newer = True
    elif cycles.index(pr_cycle_a) == cycles.index(pr_cycle_b):
        if pr_v_a is not None and pr_v_b is not None and pr_v_a > pr_v_b:
            is_newer = True
    return is_newer


def _parse_version(version: str) -> Tuple[int, int, int, str, Optional[int]]:
    pattern = r"(\d+)\.(\d+)\.(\d+)(?:([abrc]+)(\d+))?"
    match = re.match(pattern, version)
    if not match:
        raise ValueError(f"Could not parse version {version}")
    major, minor, micro, pr_cycle, pr_version = match.groups()
    return int(major), int(minor), int(micro), pr_cycle, int(pr_version) if pr_version else None


def _format_version(major: int, minor: int, micro: int, pr_cycle: Optional[str], pr_version: Optional[int]) -> str:
    return f"{major}.{minor}.{micro}{pr_cycle or ''}{pr_version or ''}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--package", required=True)
    parser.add_argument("-v", "--version", required=True)
    args = parser.parse_args()

    version_exists = check_if_version_exists(args.package, args.version)

    if version_exists:
        print("yes")  # noqa: T201
    else:
        print("no")  # noqa: T201



from __future__ import annotations

import json
from abc import abstractmethod
from collections import UserList
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    SupportsIndex,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from cognite.client import utils
from cognite.client.exceptions import CogniteMissingClientError
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._pandas_helpers import convert_nullable_int_cols, notebook_display_with_fallback
from cognite.client.utils._text import convert_all_keys_to_camel_case, to_snake_case
from cognite.client.utils._time import convert_time_attributes_to_datetime

if TYPE_CHECKING:
    import pandas

    from cognite.client import CogniteClient

EXCLUDE_VALUE = [None]


def basic_instance_dump(obj: Any, camel_case: bool) -> Dict[str, Any]:
    # TODO: Consider using inheritance?
    dumped = {k: v for k, v in vars(obj).items() if v not in EXCLUDE_VALUE and not k.startswith("_")}
    if camel_case:
        return convert_all_keys_to_camel_case(dumped)
    return dumped


class CogniteResponse:
    def __str__(self) -> str:
        item = convert_time_attributes_to_datetime(self.dump())
        return json.dumps(item, default=utils._auxiliary.json_dump_default, indent=4)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: Any) -> bool:
        return type(other) == type(self) and other.dump() == self.dump()

    def __getattribute__(self, item: Any) -> Any:
        attr = super().__getattribute__(item)
        if item == "_cognite_client" and attr is None:
            raise CogniteMissingClientError
        return attr

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance.
        """
        return basic_instance_dump(self, camel_case=camel_case)

    @classmethod
    def _load(cls, api_response: Dict[str, Any]) -> CogniteResponse:
        raise NotImplementedError

    def to_pandas(self) -> pandas.DataFrame:
        raise NotImplementedError


T_CogniteResponse = TypeVar("T_CogniteResponse", bound=CogniteResponse)


class CogniteResource:
    _cognite_client: Any

    def __new__(cls, *args: Any, **kwargs: Any) -> CogniteResource:
        obj = super().__new__(cls)
        obj._cognite_client = None
        if "cognite_client" in kwargs:
            obj._cognite_client = kwargs["cognite_client"]
        return obj

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.dump() == other.dump()

    def __str__(self) -> str:
        item = convert_time_attributes_to_datetime(self.dump())
        return json.dumps(item, default=utils._auxiliary.json_dump_default, indent=4)

    def __getattribute__(self, item: Any) -> Any:
        attr = super().__getattribute__(item)
        if item == "_cognite_client" and attr is None:
            raise CogniteMissingClientError
        return attr

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance.
        """
        return basic_instance_dump(self, camel_case=camel_case)

    @classmethod
    def _load(
        cls: Type[T_CogniteResource], resource: Union[Dict, str], cognite_client: Optional[CogniteClient] = None
    ) -> T_CogniteResource:
        if isinstance(resource, str):
            return cls._load(json.loads(resource), cognite_client=cognite_client)
        elif isinstance(resource, Dict):
            instance = cls(cognite_client=cognite_client)
            for key, value in resource.items():
                snake_case_key = to_snake_case(key)
                if hasattr(instance, snake_case_key):
                    setattr(instance, snake_case_key, value)
            return instance
        raise TypeError(f"Resource must be json str or dict, not {type(resource)}")

    def to_pandas(
        self, expand: Sequence[str] = ("metadata",), ignore: Optional[List[str]] = None, camel_case: bool = False
    ) -> pandas.DataFrame:
        """Convert the instance into a pandas DataFrame.

        Args:
            expand (List[str]): List of row keys to expand, only works if the value is a Dict.
                Will expand metadata by default.
            ignore (List[str]): List of row keys to not include when converting to a data frame.
            camel_case (bool): Convert column names to camel case (e.g. `externalId` instead of `external_id`)

        Returns:
            pandas.DataFrame: The dataframe.
        """
        ignore = [] if ignore is None else ignore
        pd = cast(Any, utils._auxiliary.local_import("pandas"))
        dumped = self.dump(camel_case=camel_case)

        for element in ignore:
            del dumped[element]
        for key in expand:
            if key in dumped:
                if isinstance(dumped[key], dict):
                    dumped.update(dumped.pop(key))
                else:
                    raise AssertionError(f"Could not expand attribute '{key}'")
        df = pd.DataFrame(columns=["value"])
        for name, value in dumped.items():
            df.loc[name] = [value]
        return df

    def _repr_html_(self) -> str:
        return notebook_display_with_fallback(self)


T_CogniteResource = TypeVar("T_CogniteResource", bound=CogniteResource)


class CognitePropertyClassUtil:
    @staticmethod
    def declare_property(schema_name: str) -> property:
        return (
            property(lambda s: s[schema_name] if schema_name in s else None)
            .setter(lambda s, v: CognitePropertyClassUtil._property_setter(s, schema_name, v))
            .deleter(lambda s: s.pop(schema_name, None))
        )

    @staticmethod
    def _property_setter(self: Any, schema_name: str, value: Any) -> None:
        if value is None:
            self.pop(schema_name, None)
        else:
            self[schema_name] = value


class CogniteResourceList(UserList, Generic[T_CogniteResource]):
    _RESOURCE: Type[CogniteResource]

    def __init__(self, resources: Collection[Any], cognite_client: Optional[CogniteClient] = None):
        for resource in resources:
            if not isinstance(resource, self._RESOURCE):
                raise TypeError(
                    f"All resources for class '{self.__class__.__name__}' must be of type "
                    f"'{self._RESOURCE.__name__}', not '{type(resource)}'."
                )
        self._cognite_client = cast("CogniteClient", cognite_client)
        super().__init__(resources)
        self._id_to_item, self._external_id_to_item = {}, {}
        if self.data:
            if hasattr(self.data[0], "external_id"):
                self._external_id_to_item = {
                    item.external_id: item for item in self.data if item.external_id is not None
                }
            if hasattr(self.data[0], "id"):
                self._id_to_item = {item.id: item for item in self.data if item.id is not None}

    def __getattribute__(self, item: Any) -> Any:
        attr = super().__getattribute__(item)
        if item == "_cognite_client" and attr is None:
            raise CogniteMissingClientError
        return attr

    def pop(self, i: int = -1) -> T_CogniteResource:
        return super().pop(i)

    def __iter__(self) -> Iterator[T_CogniteResource]:
        return super().__iter__()

    @overload
    def __getitem__(self, item: SupportsIndex) -> T_CogniteResource:
        ...

    @overload
    def __getitem__(self, item: slice) -> CogniteResourceList[T_CogniteResource]:
        ...

    def __getitem__(self, item: Any) -> Any:
        value = super().__getitem__(item)
        if isinstance(item, slice):
            c = None
            if super().__getattribute__("_cognite_client") is not None:
                c = self._cognite_client
            return self.__class__(value, cognite_client=c)
        return value

    def __str__(self) -> str:
        item = convert_time_attributes_to_datetime(self.dump())
        return json.dumps(item, default=utils._auxiliary.json_dump_default, indent=4)

    # TODO: We inherit a lot from UserList that we don't actually support...
    def extend(self, other: Collection[Any]) -> None:  # type: ignore [override]
        other_res_list = type(self)(other)  # See if we can accept the types
        if set(self._id_to_item).isdisjoint(other_res_list._id_to_item):
            super().extend(other)
            self._external_id_to_item.update(other_res_list._external_id_to_item)
            self._id_to_item.update(other_res_list._id_to_item)
        else:
            raise ValueError("Unable to extend as this would introduce duplicates")

    def dump(self, camel_case: bool = False) -> List[Dict[str, Any]]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            List[Dict[str, Any]]: A list of dicts representing the instance.
        """
        return [resource.dump(camel_case) for resource in self.data]

    def get(self, id: Optional[int] = None, external_id: Optional[str] = None) -> Optional[T_CogniteResource]:
        """Get an item from this list by id or exernal_id.

        Args:
            id (int): The id of the item to get.
            external_id (str): The external_id of the item to get.

        Returns:
            Optional[CogniteResource]: The requested item
        """
        IdentifierSequence.load(id, external_id).assert_singleton()
        if id:
            return self._id_to_item.get(id)
        return self._external_id_to_item.get(external_id)

    def to_pandas(
        self,
        camel_case: bool = False,
        expand_metadata: bool = False,
        metadata_prefix: str = "metadata.",
    ) -> pandas.DataFrame:
        """Convert the instance into a pandas DataFrame. Note that if the metadata column is expanded and there are
        keys in the metadata that already exist in the DataFrame, then an error will be raised by pd.join.

        Args:
            camel_case (bool): Convert column names to camel case (e.g. `externalId` instead of `external_id`)
            expand_metadata (bool): Expand the metadata column into separate columns.
            metadata_prefix (str): Prefix to use for metadata columns.

        Returns:
            pandas.DataFrame: The Cognite resource as a dataframe.
        """
        pd = cast(Any, utils._auxiliary.local_import("pandas"))
        df = pd.DataFrame(self.dump(camel_case=camel_case))
        df = convert_nullable_int_cols(df, camel_case)

        if expand_metadata and "metadata" in df.columns:
            # Equivalent to pd.json_normalize(df["metadata"]) but is a faster implementation.
            meta_series = df.pop("metadata").dropna()
            meta_df = pd.DataFrame(meta_series.values.tolist(), index=meta_series.index).add_prefix(metadata_prefix)
            df = df.join(meta_df)
        return df

    def _repr_html_(self) -> str:
        return notebook_display_with_fallback(self)

    @classmethod
    def _load(
        cls: Type[T_CogniteResourceList],
        resource_list: Union[List, str],
        cognite_client: Optional[CogniteClient] = None,
    ) -> T_CogniteResourceList:
        if isinstance(resource_list, str):
            return cls._load(json.loads(resource_list), cognite_client=cognite_client)
        elif isinstance(resource_list, List):
            resources = [cls._RESOURCE._load(resource, cognite_client=cognite_client) for resource in resource_list]
            return cls(resources, cognite_client=cognite_client)


T_CogniteResourceList = TypeVar("T_CogniteResourceList", bound=CogniteResourceList)


@dataclass
class PropertySpec:
    name: str
    is_container: bool = False
    is_nullable: bool = True


class CogniteUpdate:
    def __init__(self, id: Optional[int] = None, external_id: Optional[str] = None):
        self._id = id
        self._external_id = external_id
        self._update_object: Dict[str, Any] = {}

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.dump() == other.dump()

    def __str__(self) -> str:
        return json.dumps(self.dump(), default=utils._auxiliary.json_dump_default, indent=4)

    def __repr__(self) -> str:
        return str(self)

    def _set(self, name: str, value: Any) -> None:
        update_obj = self._update_object.get(name, {})
        assert (
            "remove" not in update_obj and "add" not in update_obj
        ), "Can not call set after adding or removing fields from an update object."
        self._update_object[name] = {"set": value}

    def _set_null(self, name: str) -> None:
        self._update_object[name] = {"setNull": True}

    def _add(self, name: str, value: Any) -> None:
        update_obj = self._update_object.get(name, {})
        assert "set" not in update_obj, "Can not call remove or add fields after calling set on an update object."
        assert (
            "add" not in update_obj
        ), "Can not call add twice on the same object, please combine your objects and pass them to add in one call."
        update_obj["add"] = value
        self._update_object[name] = update_obj

    def _remove(self, name: str, value: Any) -> None:
        update_obj = self._update_object.get(name, {})
        assert "set" not in update_obj, "Can not call remove or add fields after calling set on an update object."
        assert (
            "remove" not in update_obj
        ), "Can not call remove twice on the same object, please combine your items and pass them to remove in one call."
        update_obj["remove"] = value
        self._update_object[name] = update_obj

    def _modify(self, name: str, value: Any) -> None:
        update_obj = self._update_object.get(name, {})
        assert "set" not in update_obj, "Can not call remove or add fields after calling set on an update object."
        assert (
            "modify" not in update_obj
        ), "Can not call modify twice on the same object, please combine your items and pass them to modify in one call."
        update_obj["modify"] = value
        self._update_object[name] = update_obj

    def dump(self, camel_case: bool = True) -> Dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance.
        """
        assert camel_case is True, "snake_case is currently unsupported"  # TODO maybe (when unifying classes)
        dumped: Dict[str, Any] = {"update": self._update_object}
        if self._id is not None:
            dumped["id"] = self._id
        elif self._external_id is not None:
            dumped["externalId"] = self._external_id
        return dumped

    @classmethod
    @abstractmethod
    def _get_update_properties(cls) -> list[PropertySpec]:
        raise NotImplementedError


T_CogniteUpdate = TypeVar("T_CogniteUpdate", bound=CogniteUpdate)


class CognitePrimitiveUpdate(Generic[T_CogniteUpdate]):
    def __init__(self, update_object: T_CogniteUpdate, name: str):
        self._update_object = update_object
        self._name = name

    def _set(self, value: Union[None, str, int, bool]) -> T_CogniteUpdate:
        if value is None:
            self._update_object._set_null(self._name)
        else:
            self._update_object._set(self._name, value)
        return self._update_object


class CogniteObjectUpdate(Generic[T_CogniteUpdate]):
    def __init__(self, update_object: T_CogniteUpdate, name: str) -> None:
        self._update_object = update_object
        self._name = name

    def _set(self, value: Dict) -> T_CogniteUpdate:
        self._update_object._set(self._name, value)
        return self._update_object

    def _add(self, value: Dict) -> T_CogniteUpdate:
        self._update_object._add(self._name, value)
        return self._update_object

    def _remove(self, value: List) -> T_CogniteUpdate:
        self._update_object._remove(self._name, value)
        return self._update_object


class CogniteListUpdate(Generic[T_CogniteUpdate]):
    def __init__(self, update_object: T_CogniteUpdate, name: str):
        self._update_object = update_object
        self._name = name

    def _set(self, value: List) -> T_CogniteUpdate:
        self._update_object._set(self._name, value)
        return self._update_object

    def _add(self, value: List) -> T_CogniteUpdate:
        self._update_object._add(self._name, value)
        return self._update_object

    def _remove(self, value: List) -> T_CogniteUpdate:
        self._update_object._remove(self._name, value)
        return self._update_object

    def _modify(self, value: List) -> T_CogniteUpdate:
        self._update_object._modify(self._name, value)
        return self._update_object


class CogniteLabelUpdate(Generic[T_CogniteUpdate]):
    def __init__(self, update_object: T_CogniteUpdate, name: str) -> None:
        self._update_object = update_object
        self._name = name

    def _set(self, external_ids: Union[str, List[str]]) -> T_CogniteUpdate:
        self._update_object._set(self._name, self._wrap_ids(self._wrap_in_list(external_ids)))
        return self._update_object

    def _add(self, external_ids: Union[str, List[str]]) -> T_CogniteUpdate:
        self._update_object._add(self._name, self._wrap_ids(self._wrap_in_list(external_ids)))
        return self._update_object

    def _remove(self, external_ids: Union[str, List[str]]) -> T_CogniteUpdate:
        self._update_object._remove(self._name, self._wrap_ids(self._wrap_in_list(external_ids)))
        return self._update_object

    def _wrap_ids(self, external_ids: List[str]) -> List[Dict[str, str]]:
        return [{"externalId": external_id} for external_id in external_ids]

    def _wrap_in_list(self, external_ids: Union[str, List[str]]) -> List[str]:
        return external_ids if isinstance(external_ids, list) else [external_ids]


class CogniteFilter:
    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.dump() == other.dump()

    def __str__(self) -> str:
        item = convert_time_attributes_to_datetime(self.dump())
        return json.dumps(item, default=utils._auxiliary.json_dump_default, indent=4)

    def __repr__(self) -> str:
        return str(self)

    def __getattribute__(self, item: Any) -> Any:
        attr = super().__getattribute__(item)
        if item == "_cognite_client" and attr is None:
            raise CogniteMissingClientError
        return attr

    @classmethod
    def _load(cls: Type[T_CogniteFilter], resource: Union[Dict, str]) -> T_CogniteFilter:
        if isinstance(resource, str):
            return cls._load(json.loads(resource))
        elif isinstance(resource, Dict):
            instance = cls()
            for key, value in resource.items():
                snake_case_key = to_snake_case(key)
                if hasattr(instance, snake_case_key):
                    setattr(instance, snake_case_key, value)
            return instance
        raise TypeError(f"Resource must be json str or dict, not {type(resource)}")

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance.
        """
        return basic_instance_dump(self, camel_case=camel_case)


T_CogniteFilter = TypeVar("T_CogniteFilter", bound=CogniteFilter)



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from collections import UserDict
from dataclasses import dataclass, field
from typing import Any, Dict, Literal, Mapping, Optional, Type, cast

from cognite.client.data_classes.data_modeling.ids import ViewId
from cognite.client.data_classes.data_modeling.instances import (
    EdgeList,
    EdgeListWithCursor,
    InstanceSort,
    NodeList,
    NodeListWithCursor,
    PropertyValue,
)
from cognite.client.data_classes.filters import Filter
from cognite.client.utils._auxiliary import local_import


@dataclass
class SourceSelector:
    source: ViewId
    properties: list[str]

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        return {
            "source": self.source.dump(camel_case),
            "properties": self.properties,
        }

    @classmethod
    def load(cls, data: dict | str) -> SourceSelector:
        data = json.loads(data) if isinstance(data, str) else data
        return cls(
            source=ViewId.load(data["source"]),
            properties=data["properties"],
        )


@dataclass
class Select:
    sources: list[SourceSelector] = field(default_factory=list)
    sort: list[InstanceSort] = field(default_factory=list)
    limit: Optional[int] = None

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        output: Dict[str, Any] = {}
        if self.sources:
            output["sources"] = [
                {"source": source.source.dump(camel_case), "properties": source.properties} for source in self.sources
            ]
        if self.sort:
            output["sort"] = [s.dump(camel_case) for s in self.sort]
        if self.limit:
            output["limit"] = self.limit
        return output

    @classmethod
    def load(cls, data: dict | str) -> Select:
        data = json.loads(data) if isinstance(data, str) else data
        return cls(
            sources=[SourceSelector.load(source) for source in data.get("sources", [])],
            sort=[InstanceSort._load(s) for s in data.get("sort", [])],
            limit=data.get("limit"),
        )


class Query:
    """Query allows you to do advanced queries on the data model.

    Args:
        with\_ (dict[str, ResultSetExpression]): A dictionary of result set expressions to use in the query. The keys
                                                are used to reference the result set expressions in the select and
                                                parameters.
        select (dict[str, Select]): A dictionary of select expressions to use in the query. The keys must match the
                                    keys in the with\_ dictionary. The select expressions define which properties to
                                    include in the result set.
        parameters (Optional[dict[str, PropertyValue]]): Values in filters can be parameterised. Parameters are provided
                                                         as part of the query object, and referenced in the filter itself.
        cursors (Optional[Mapping[str, Optional[str]]]): A dictionary of cursors to use in the query. These are for
                                                         pagination purposes, for example, in the sync endpoint.

    """

    def __init__(
        self,
        with_: dict[str, ResultSetExpression],
        select: dict[str, Select],
        parameters: Optional[dict[str, PropertyValue]] = None,
        cursors: Optional[Mapping[str, Optional[str]]] = None,
    ) -> None:
        with_keys = set(with_)
        if not_matching := set(select) - with_keys:
            raise ValueError(
                f"The select keys must match the with keys, the following are not matching: {not_matching}"
            )

        self.with_ = with_
        self.select = select
        self.parameters = parameters
        self.cursors = cursors or {k: None for k in select}

    def instance_type_by_result_expression(self) -> dict[str, Type[NodeList] | Type[EdgeList]]:
        return {k: NodeList if isinstance(v, NodeResultSetExpression) else EdgeList for k, v in self.with_.items()}

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        output: Dict[str, Any] = {
            "with": {k: v.dump(camel_case) for k, v in self.with_.items()},
            "select": {k: v.dump(camel_case) for k, v in self.select.items()},
        }
        if self.parameters:
            output["parameters"] = dict(self.parameters.items())
        if self.cursors:
            output["cursors"] = dict(self.cursors.items())
        return output

    @classmethod
    def load_yaml(cls, data: str) -> Query:
        yaml = cast(Any, local_import("yaml"))
        return cls.load(yaml.safe_load(data))

    @classmethod
    def load(cls, data: str | dict[str, Any]) -> Query:
        data = json.loads(data) if isinstance(data, str) else data

        if not (with_ := data.get("with")):
            raise ValueError("The query must contain a with key")

        loaded: Dict[str, Any] = {"with_": {k: ResultSetExpression.load(v) for k, v in with_.items()}}
        if not (select := data.get("select")):
            raise ValueError("The query must contain a select key")
        loaded["select"] = {k: Select.load(v) for k, v in select.items()}

        if parameters := data.get("parameters"):
            loaded["parameters"] = dict(parameters.items())
        if cursors := data.get("cursors"):
            loaded["cursors"] = dict(cursors.items())
        return cls(**loaded)

    def __eq__(self, other: Any) -> bool:
        return type(other) == type(self) and self.dump() == other.dump()


class ResultSetExpression(ABC):
    @abstractmethod
    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        ...

    @classmethod
    def load(cls, query: dict[str, Any]) -> ResultSetExpression:
        if "sort" in query:
            sort = [InstanceSort(**sort) for sort in query["sort"]]
        else:
            sort = []

        if "nodes" in query:
            query_node = query["nodes"]
            node = {
                "from_": query_node.get("from"),
                "filter": Filter.load(query_node["filter"]) if "filter" in query_node else None,
            }
            return NodeResultSetExpression(sort=sort, limit=query.get("limit"), **node)
        elif "edges" in query:
            query_edge = query["edges"]
            edge = {
                "from_": query_edge.get("from"),
                "max_distance": query_edge.get("maxDistance"),
                "direction": query_edge.get("direction"),
                "filter": Filter.load(query_edge["filter"]) if "filter" in query_edge else None,
                "node_filter": Filter.load(query_edge["nodeFilter"]) if "nodeFilter" in query_edge else None,
                "termination_filter": Filter.load(query_edge["terminationFilter"])
                if "terminationFilter" in query_edge
                else None,
                "limit_each": query_edge.get("limitEach"),
            }
            return EdgeResultSetExpression(**edge, sort=sort, limit=query.get("limit"))
        else:
            raise NotImplementedError(f"Unknown query type: {query}")

    def __eq__(self, other: Any) -> bool:
        return type(other) == type(self) and self.dump() == other.dump()


class NodeResultSetExpression(ResultSetExpression):
    def __init__(
        self,
        from_: Optional[str] = None,
        filter: Optional[Filter] = None,
        sort: Optional[list[InstanceSort]] = None,
        limit: Optional[int] = None,
    ):
        self.from_ = from_
        self.filter = filter
        self.sort = sort
        self.limit = limit

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        output: Dict[str, Any] = {"nodes": {}}
        nodes = output["nodes"]
        if self.from_:
            nodes["from"] = self.from_
        if self.filter:
            nodes["filter"] = self.filter.dump()

        if self.sort:
            output["sort"] = [s.dump(camel_case=camel_case) for s in self.sort]
        if self.limit:
            output["limit"] = self.limit

        return output


class EdgeResultSetExpression(ResultSetExpression):
    def __init__(
        self,
        from_: Optional[str] = None,
        max_distance: Optional[int] = None,
        direction: Literal["outwards", "inwards"] = "outwards",
        filter: Optional[Filter] = None,
        node_filter: Optional[Filter] = None,
        termination_filter: Optional[Filter] = None,
        limit_each: Optional[int] = None,
        sort: Optional[list[InstanceSort]] = None,
        post_sort: Optional[list[InstanceSort]] = None,
        limit: Optional[int] = None,
    ):
        self.from_ = from_
        self.max_distance = max_distance
        self.direction = direction
        self.filter = filter
        self.node_filter = node_filter
        self.termination_filter = termination_filter
        self.limit_each = limit_each
        self.sort = sort
        self.post_sort = post_sort
        self.limit = limit

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        output: Dict[str, Any] = {"edges": {}}
        edges = output["edges"]
        if self.from_:
            edges["from"] = self.from_
        if self.max_distance:
            edges["maxDistance" if camel_case else "max_distance"] = self.max_distance
        if self.direction:
            edges["direction"] = self.direction
        if self.filter:
            edges["filter"] = self.filter.dump()
        if self.node_filter:
            edges["nodeFilter" if camel_case else "node_filter"] = self.node_filter.dump()
        if self.termination_filter:
            edges["terminationFilter" if camel_case else "termination_filter"] = self.termination_filter.dump()
        if self.limit_each:
            edges["limitEach" if camel_case else "limit_each"] = self.limit_each

        if self.sort:
            output["sort"] = [s.dump(camel_case=camel_case) for s in self.sort]
        if self.post_sort:
            output["postSort" if camel_case else "post_sort"] = [s.dump(camel_case=camel_case) for s in self.post_sort]
        if self.limit:
            output["limit"] = self.limit

        return output


class QueryResult(UserDict):
    def __getitem__(self, item: str) -> NodeListWithCursor | EdgeListWithCursor:
        return super().__getitem__(item)

    @property
    def cursors(self) -> dict[str, str]:
        return {key: value.cursor for key, value in self.items()}

    @classmethod
    def load(
        cls,
        data: dict[str, Any] | str,
        default_by_reference: dict[str, Type[NodeList] | Type[EdgeList]],
        cursors: dict[str, Any] | None = None,
    ) -> QueryResult:
        data = json.loads(data) if isinstance(data, str) else data
        instance = cls()
        for key, values in data.items():
            if not values:
                instance[key] = default_by_reference[key]([])
            elif values[0].get("instanceType") == "node":
                instance[key] = NodeListWithCursor._load(values)
                if cursors:
                    instance[key].cursor = cursors.get(key)
            elif values[0].get("instanceType") == "edge":
                instance[key] = EdgeListWithCursor._load(values)
                if cursors:
                    instance[key].cursor = cursors.get(key)
            else:
                raise ValueError(f"Unexpected instance type {values[0].get('instanceType')}")

        return instance

    def get_nodes(self, result_expression: str) -> NodeListWithCursor:
        result = super().__getitem__(result_expression)
        if isinstance(result, NodeListWithCursor):
            return result
        raise RuntimeError(
            f"{result_expression} is not a {NodeListWithCursor.__name__}. Try {self.get_edges.__name__} instead."
        )

    def get_edges(self, result_expression: str) -> EdgeListWithCursor:
        result = super().__getitem__(result_expression)
        if isinstance(result, EdgeListWithCursor):
            return result
        raise RuntimeError(
            f"{result_expression} is not a {EdgeListWithCursor.__name__}. Try {self.get_nodes.__name__} instead."
        )


# class SetOperationResultSetExpression(ResultSetExpression):
#     ...
#
#
# class UnionAllTableExpression(SetOperationResultSetExpression):
#     def __init__(
#         self, union_all: list[QuerySetOperationTableExpression | str], except_: list[str] = None, limit: int = None
#     ):
#         self.union_all = union_all
#         self.except_ = except_
#         self.limit = limit
#
#
# class UnionTableExpression(SetOperationResultSetExpression):
#     def __init__(
#         self, union: list[QuerySetOperationTableExpression | str], except_: list[str] = None, limit: int = None
#     ):
#         self.union = union
#         self.except_ = except_
#         self.limit = limit
#
#
# class IntersectTableExpression(SetOperationResultSetExpression):
#     def __init__(
#         self, intersection: list[QuerySetOperationTableExpression | str], except_: list[str] = None, limit: int = None
#     ):
#         self.intersect = intersection
#         self.except_ = except_
#         self.limit = limit



from __future__ import annotations

from typing import Any, Optional

from cognite.client._api_client import APIClient
from cognite.client.data_classes.data_modeling import DataModelIdentifier
from cognite.client.data_classes.data_modeling.graphql import DMLApplyResult
from cognite.client.data_classes.data_modeling.ids import DataModelId
from cognite.client.exceptions import CogniteGraphQLError, GraphQLErrorSpec


class DataModelingGraphQLAPI(APIClient):
    def _post_graphql(self, url_path: str, query: str) -> dict[str, Any]:
        res = self._post(url_path=url_path, json={"query": query})
        json_res = res.json()
        if (errors := json_res.get("errors")) is not None:
            raise CogniteGraphQLError([GraphQLErrorSpec.load(error) for error in errors])
        return json_res["data"]

    def apply_dml(
        self,
        id: DataModelIdentifier,
        dml: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        previous_version: Optional[str] = None,
    ) -> DMLApplyResult:
        """Apply the DML for a given data model.

        Args:
            id (DataModelIdentifier): The data model to apply DML to.
            dml (str): The DML to apply.
            previous_version (Optional[str]): The previous version of the data model. Specify to reuse view versions from
                previous data model version.
            name (Optional[str]): The name of the data model.
            description (Optional[str]): The description of the data model.

        Returns:
            DMLApplyResult: The id of the updated data model.

        Examples:

            Apply DML::

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.data_modeling.graphql.apply_dml(
                ...     ("mySpace", "myDataModel", "v1"),
                ...     dml="type MyType { id: String! }"
                ... )
        """
        data_model_id = DataModelId.load(id)
        graphql_body = f"""
            mutation UpsertGraphQlDmlVersion{{
                upsertGraphQlDmlVersion(
                    graphQlDmlVersion: {{
                        space: "{data_model_id.space}",
                        externalId: "{data_model_id.external_id}",
                        version: "{data_model_id.version}",
                        graphQlDml: "{dml}",
                        previousVersion: "{previous_version}",
                        name: "{name}",
                        description: "{description}"
                    }}
                ) {{
                    result {{
                        space
                        externalId
                        version
                        name
                        description
                        createdTime
                        lastUpdatedTime
                    }}
                }}
            }}
        """
        res = self._post_graphql(url_path="/dml/graphql", query=graphql_body)
        return DMLApplyResult.load(res["upsertGraphQlDmlVersion"]["result"])
